#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEServer.h>
#include <BLE2902.h>
#include <AccelStepper.h>

// Stepper motor setup
const int stepPin1 = 18;
const int dirPin1 = 19;
AccelStepper stepper1(1, stepPin1, dirPin1);

// BLE Characteristics and Service UUIDs
#define SERVICE_UUID "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_RX "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"
#define CHARACTERISTIC_UUID_TX "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"

// Global variables
BLECharacteristic *pTxCharacteristic;
BLECharacteristic *pRxCharacteristic;
bool deviceConnected = false;
String btData;

// Variables for motor control
int currentAngle = 0;  // Start from 0 degrees
int increment = 0;     // Increment received via BLE
bool rotationComplete = false;  // Flag for completing 360°

class MyServerCallbacks : public BLEServerCallbacks {
    void onConnect(BLEServer *pServer) {
        deviceConnected = true;
        Serial.println("Device connected");
    }

    void onDisconnect(BLEServer *pServer) {
        deviceConnected = false;
        Serial.println("Device disconnected");
        pServer->startAdvertising(); // Restart advertising after disconnection
    }
};

class MyCharacteristicCallbacks : public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) override {
        Serial.println("onWrite triggered"); // Debug print
        String btData = String(pCharacteristic->getValue().c_str()); // Convert std::string to Arduino String
        btData.trim(); // Remove any whitespace or extraneous characters
        if (btData.length() > 0) {
            Serial.println("Received: " + btData);

            // Parse and process data if it is enclosed in square brackets
            if (btData[0] == '[' && btData[btData.length() - 1] == ']') {
                btData = btData.substring(1, btData.length() - 1); // Remove square brackets
                increment = btData.toInt(); // Convert to integer
                Serial.println("Valid increment received: " + String(increment));

                if (!rotationComplete) {
                    // Calculate new target angle
                    while (currentAngle < 360) {
                        currentAngle += increment;
                        if (currentAngle > 360) break; // Avoid overshooting 360

                        // Map angle to stepper position
                        long newPosition = map(currentAngle, 0, 360, 0, 818);

                        // Move to new position
                        stepper1.runToNewPosition(newPosition);

                        // Wait for 3 seconds at the stop
                        delay(3000);
                        Serial.println("Reached the next stop!");
                    }

                    // Mark rotation as complete
                    if (currentAngle >= 360) {
                        rotationComplete = true;
                        Serial.println("Completed full rotation. Waiting for new command...");
                        stepper1.setCurrentPosition(0); // Reset stepper motor position
                        currentAngle = 0; // Reset for next command
                    }
                } else {
                    // Reset flag and allow new rotation
                    rotationComplete = false;
                }
            } else {
                Serial.println("Invalid message format. Please send data in square brackets [ ].");
            }
        }
    }
};

void setup() {
    Serial.begin(115200);
    Serial.println("Initializing BLE...");

    // Initialize BLE
    BLEDevice::init("ESP32_Motor_Controller"); // Set the device name
    BLEServer *pServer = BLEDevice::createServer();
    pServer->setCallbacks(new MyServerCallbacks());

    // Create the BLE Service
    BLEService *pService = pServer->createService(SERVICE_UUID);

    // Create BLE TX Characteristic
    pTxCharacteristic = pService->createCharacteristic(
        CHARACTERISTIC_UUID_TX,
        BLECharacteristic::PROPERTY_NOTIFY
    );
    pTxCharacteristic->addDescriptor(new BLE2902());

    // Create BLE RX Characteristic
    pRxCharacteristic = pService->createCharacteristic(
        CHARACTERISTIC_UUID_RX,
        BLECharacteristic::PROPERTY_READ|
        BLECharacteristic::PROPERTY_WRITE
    );
    pRxCharacteristic->setCallbacks(new MyCharacteristicCallbacks());

    // Start the service and advertising
    pService->start();
    pServer->getAdvertising()->start();
    Serial.println("BLE Advertising Started. Waiting for connections...");

    // Stepper motor setup
    stepper1.setCurrentPosition(0);
    stepper1.setMaxSpeed(150); // Maximum speed in steps per second
    stepper1.setAcceleration(400); // Set acceleration in steps per second^2
//    Serial.println("Stepper motor initialized.");
}

void loop() {
    // Handle BLE disconnections
    if (!deviceConnected && rotationComplete) {
        // Reset the system for a new connection
        rotationComplete = false;
        currentAngle = 0;
        stepper1.setCurrentPosition(0);
        Serial.println("System reset. Waiting for new BLE connection...");
    }

    // Optional: Add a small delay to reduce CPU usage
    delay(10);
}
